/* @(#) DirEnt.c 1995-07-27 - dir entries via Win32.
 * Copyright (c) 1995 by JR (John Rogers).
 *
 * AUTHOR - JR (John Rogers, 72634.2402@CompuServe.com)
 */

/*lint -e506 */ /* Allow "Constant value Boolean". */
#include <assert.h>     /* assert(). */
#include <dirent.h>     /* struct dirent, opendir(). */
#include <errno.h>      /* errno. */
#include <stdio.h>      /* fprintf(), stderr. */
#include <stdlib.h>     /* NULL. */
#include <string.h>     /* strcpy(), strlen(). */
#include <windows.h>    /* FindFirstFile(), etc. */

/*EVALTWICE*/
#define IS_DIR_SEP(ch) \
   ( ((ch) == '/') || ((ch) == ':') || (ch == '\\') )

#define MY_SIGNATURE 0xFF112233UL

#ifndef NAME_MAX
#define NAME_MAX MAX_PATH
#endif

struct tagDIR {
   unsigned long    Signature;
   HANDLE           Win32_Find_Handle;
   WIN32_FIND_DATA  Found_File_Data;
   BOOL             Unparsed_Data_Available;

   char             Search_Path[ MAX_PATH ];

   struct dirent    Parsed_File_Data;
   /* This is followed by the current file name. */
};

static void
Extract_Data_For_One(
   struct dirent *    Dest,
   LPWIN32_FIND_DATA  Src)
{
   assert( Dest != NULL );
   assert( Src != NULL );

   (void) strcpy( Dest->d_name, Src->cFileName );
   Dest->d_namlen = strlen( Dest->d_name );
}

static void
Record_Last_Win32_Error(
   void)
{
   const DWORD  Win32_Error_Code = GetLastError();

   assert( Win32_Error_Code != (DWORD) 0 );

   /* convert Win32 to C error code. */
   switch( Win32_Error_Code ) {
   case ERROR_ACCESS_DENIED:
      errno = EACCES;
      break;
   case ERROR_BAD_NETPATH:
      errno = ENOENT;
      break;
   case ERROR_BAD_PATHNAME:  /* This is guess! */
      errno = ENOTDIR;
      break;
   case ERROR_DIRECTORY:
      errno = ENOTDIR;
      break;
   case ERROR_FILENAME_EXCED_RANGE:
      errno = ENAMETOOLONG;
      break;
   case ERROR_INVALID_NAME:
      errno = ENOENT;
      break;
   case ERROR_PATH_NOT_FOUND:
      errno = ENOENT;
      break;
   case ERROR_TOO_MANY_OPEN_FILES:
      errno = EMFILE;  /* Or use ENFILE here? */
      break;
   default:

      (void) fprintf(
            stderr,
            "dirent routines got unexpected Win32 "
            "error code %ld.\n",
            (unsigned long) Win32_Error_Code );
      errno = EINVAL;  /* wild guess */
   }
   assert( errno != 0 );
}

int
closedir(
   DIR *  Private_Data)
{
   if ( (Private_Data == NULL)
      || (Private_Data->Signature != MY_SIGNATURE) ) {
      errno = EBADF;
      return (-1);  /* indicate error */
   }

   /* Close Win32 find handle (unless empty dir). */
   if (Private_Data->Win32_Find_Handle
         != INVALID_HANDLE_VALUE) {
      if (!FindClose(
            Private_Data->Win32_Find_Handle)) {
         Record_Last_Win32_Error();
         return (-1);  /* indicate error */
      }
   }

   free( Private_Data );

   return (0);  /* indicate success */
}

DIR *
opendir(
   const char *  Dir_Name)
{
   size_t  Dir_Name_Len;
   char    Last_Char;
   DIR *   Private_Data = NULL;
   if (Dir_Name == NULL) {
      errno = EINVAL;
      goto Error;
   } else if (*Dir_Name == '\0') {
      errno = ENOENT;
      goto Error;
   }
   
   Private_Data = malloc(
         sizeof( struct tagDIR )
         + NAME_MAX );
   if (Private_Data == NULL) {
      errno = ENOMEM;
      goto Error;
   }
   Private_Data->Signature = MY_SIGNATURE;

   Dir_Name_Len = strlen(Dir_Name);

   /*
    * Build Search_Path from Dir_Name, "/" if needed,
    * and "*".
    */
   (void) strcpy(
         Private_Data->Search_Path,     /* dest */
         Dir_Name);                     /* src */
   Last_Char = Dir_Name[Dir_Name_Len-1];
   if ( !IS_DIR_SEP( Last_Char ) ) {
      (void) strcat(
            Private_Data->Search_Path,  /* dest */
            "/");                       /* src */
   }
   (void) strcat(
         Private_Data->Search_Path,     /* dest */
         "*");                          /* src */

   /* Find first file matching that search path. */
   Private_Data->Win32_Find_Handle = FindFirstFile(
         (LPCSTR) Private_Data->Search_Path,
         &(Private_Data->Found_File_Data));

   if (Private_Data->Win32_Find_Handle
         == INVALID_HANDLE_VALUE) {

      if (GetLastError() == ERROR_FILE_NOT_FOUND) {
         /* Tell readdir() NOT to parse data. */
         Private_Data->Unparsed_Data_Available = FALSE;
         return (Private_Data);
      }
      Record_Last_Win32_Error();
      goto Error;
   }

   /* Tell readdir() that there is data to parse. */
   Private_Data->Unparsed_Data_Available = TRUE;

   return (Private_Data);

Error:  /* Assumes errno has already been set. */
   if (Private_Data != NULL) {
      free( Private_Data );
   }
   assert( errno != 0 );
   return (NULL);
}

struct dirent *
readdir(
   DIR *  Private_Data)
{
   if ( (Private_Data == NULL)
      || (Private_Data->Signature != MY_SIGNATURE) ) {
      errno = EBADF;
      return (NULL);  /* indicate error */
   }

   if ( ! (Private_Data->Unparsed_Data_Available) ) {

      /* special case empty dir */
      if (Private_Data->Win32_Find_Handle
            == INVALID_HANDLE_VALUE) {
         goto Indicate_EOF;
      }

      /* normal dir - read another */
      if ( !FindNextFile(
             Private_Data->Win32_Find_Handle,
             &(Private_Data->Found_File_Data)) ) {
         const DWORD Last_Error = GetLastError();

         if ( (Last_Error == ERROR_NO_MORE_FILES)
           || (Last_Error == ERROR_PATH_NOT_FOUND) ) {
            goto Indicate_EOF;
         } else {
            Record_Last_Win32_Error();
            return (NULL);  /* indicate error */
         }
      }
   }

   Extract_Data_For_One(
         &(Private_Data->Parsed_File_Data),  /* dest */
         &(Private_Data->Found_File_Data));  /* src */

   /* Don't reparse same data next time. */
   Private_Data->Unparsed_Data_Available = FALSE;

   return ( &(Private_Data->Parsed_File_Data) );

Indicate_EOF:  /* errno unchanged says end of file. */
   return (NULL);
}

void
rewinddir(
   DIR *  Private_Data)
{
   if ( (Private_Data == NULL)
      || (Private_Data->Signature != MY_SIGNATURE) ) {
      errno = EBADF;
      return;
   }

   if (Private_Data->Win32_Find_Handle
         == INVALID_HANDLE_VALUE)
      return;  /* Empty dir; nothing to rewind. */

      /* Special case bizarre empty directory. */
   /* Close Win32 find handle. */
   if (!FindClose(Private_Data->Win32_Find_Handle)) {
      Record_Last_Win32_Error();
      return;
   }

   Private_Data->Win32_Find_Handle = FindFirstFile(
         (LPCSTR) Private_Data->Search_Path,
         &(Private_Data->Found_File_Data));

   if (Private_Data->Win32_Find_Handle
         == INVALID_HANDLE_VALUE) {
      Record_Last_Win32_Error();
      return;
   }

   /* Tell readdir() that there is data to parse. */
   Private_Data->Unparsed_Data_Available = TRUE;
}
